"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const marked = require("marked");
const os_1 = require("os");
const path = require("path");
const tar = require("tar");
const fs_1 = require("fs");
const errors_1 = require("./errors");
const fs_2 = require("./fs");
const npm_1 = require("./npm");
/**
 * Environment variable set when the plugin is run and intended to emit
 * a bundle.
 */
exports.bundleEnv = 'MIIX_PLZ_BUNDLE';
/**
 * The BundleEmitted packages everything up nicely and emits it to the
 * provided Notifier after the webpack compilation succeeds.
 */
class BundleEmitter {
    constructor(notifier, isEnabled = !!process.env[exports.bundleEnv]) {
        this.notifier = notifier;
        this.isEnabled = isEnabled;
        /**
         * Target file to write the tarball into.
         */
        this.fileTarget = path.join(os_1.tmpdir(), `miix-bundle-${Date.now()}.tar.gz`);
    }
    /**
     * Attaches the emitted to the webpack compiler.
     */
    apply(compiler) {
        if (!this.isEnabled) {
            return;
        }
        compiler.hooks.done.tapPromise('MiixWebpackPlugin#bundleEmitter', (stats) => __awaiter(this, void 0, void 0, function* () {
            if (stats.hasErrors()) {
                return;
            }
            try {
                this.notifier.printNotification(yield this.startBundling(compiler));
            }
            catch (err) {
                try {
                    fs_1.unlinkSync(this.fileTarget);
                }
                catch (e) {
                    // ignored
                }
                this.notifier.printNotification({
                    kind: 3 /* BundleFailed */,
                    error: err.stack || err.message || err,
                });
            }
        }));
    }
    startBundling(compiler) {
        return __awaiter(this, void 0, void 0, function* () {
            const projectPath = (yield npm_1.getProjectPath(compiler.context));
            const packageJson = yield npm_1.mustLoadPackageJson(projectPath);
            const output = compiler.outputPath;
            yield this.verifyIntegrity(output);
            yield this.createTarball(projectPath, output);
            return {
                kind: 2 /* BundleCreated */,
                location: this.fileTarget,
                readme: yield this.renderReadme(projectPath),
            };
        });
    }
    /**
     * Throws an PackageIntegrityError if there are any obvious problems with
     * the bundler output.
     */
    verifyIntegrity(output) {
        return __awaiter(this, void 0, void 0, function* () {
            const home = path.join(output, 'index.html');
            if (!(yield fs_2.exists(home))) {
                throw new errors_1.PackageIntegrityError(`An index.html is missing in your project output (${home} should exist)`);
            }
        });
    }
    /**
     * Tries to render the readme in the project, returning the HTML
     * if it's able to do so.
     */
    renderReadme(projectPath) {
        return __awaiter(this, void 0, void 0, function* () {
            const readme = yield npm_1.findReadme(projectPath);
            if (!readme) {
                return null;
            }
            const contents = yield fs_2.readFile(readme);
            return marked(contents);
        });
    }
    /**
     * Compresses the output dir into the target tarball.
     */
    createTarball(projectPath, output) {
        return __awaiter(this, void 0, void 0, function* () {
            const readme = yield npm_1.findReadme(projectPath);
            if (readme) {
                yield npm_1.copy(readme, path.join(output, 'readme.md'));
            }
            yield npm_1.copy((yield npm_1.findPackageJson(projectPath)), path.join(output, 'package.json'));
            yield tar.create({
                file: this.fileTarget,
                gzip: { level: 9 },
                cwd: output,
            }, yield fs_2.readDir(output));
        });
    }
}
exports.BundleEmitter = BundleEmitter;
