"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const bundle_emitter_1 = require("./bundle-emitter");
const errors_1 = require("./errors");
const fs_1 = require("./fs");
const html_renderer_1 = require("./html-renderer");
const locale_packager_1 = require("./locale-packager");
const metadata_1 = require("./metadata/metadata");
const notifier_1 = require("./notifier");
const npm_1 = require("./npm");
const webpack_tools_1 = require("./webpack-tools");
/**
 * MixerPlugin is the webpack plugin to handle packaging Interactive output
 * to ship and develop with. Primarily, it modifies the user-provided
 * HTML page to insert the Mixer standard library and additional controls
 * in development mode.
 */
class MixerPlugin {
    constructor(options) {
        this.options = options;
        this.notifier = new notifier_1.Notifier();
        this.bundle = new bundle_emitter_1.BundleEmitter(this.notifier);
    }
    apply(compiler) {
        if (!compiler.hooks) {
            throw new Error('The cdk webpack plugin requires Webpack version 4. ' +
                'Please run "npm install --save-dev webpack@^4.0.0" to update.');
        }
        this.notifier.apply(compiler);
        this.bundle.apply(compiler);
        compiler.hooks.emit.tapPromise('MiixWebpackPlugin#plugin', (compilation) => __awaiter(this, void 0, void 0, function* () {
            const projectPath = yield npm_1.getProjectPath(compiler.context);
            if (!projectPath) {
                throw new Error('Could not find your project path, are you missing a package.json?');
            }
            const jsonPath = (yield npm_1.findPackageJson(projectPath));
            webpack_tools_1.addFilesToCompilation(compilation, jsonPath);
            const packageJson = yield npm_1.mustLoadPackageJson(projectPath);
            this.package = yield metadata_1.createPackage(packageJson, projectPath);
            this.notifier.updateMetadata(this.package);
            yield this.addProductionFiles(compiler, compilation);
        }));
    }
    addProductionFiles(compiler, compilation) {
        return __awaiter(this, void 0, void 0, function* () {
            const packager = new locale_packager_1.LocalePackager(compilation);
            const locales = yield packager.compile(this.options.locales);
            const homepagePath = path.resolve(compiler.context, this.options.homepage);
            if (!(yield fs_1.exists(homepagePath))) {
                throw new errors_1.PackageIntegrityError(`We couldn't find your homepage in ${homepagePath}. Make sure the "homepage" ` +
                    `options is configured correctly in the MixerPlugin config in your webpack.config.js`);
            }
            const result = yield new html_renderer_1.HomepageRenderer(homepagePath, this.package, Object.keys(locales)).render(compiler);
            compilation.assets['index.html'] = webpack_tools_1.contentsToAsset(result);
            webpack_tools_1.addFilesToCompilation(compilation, homepagePath);
        });
    }
    addFiles(compilation, files) {
        return __awaiter(this, void 0, void 0, function* () {
            return Promise.all(Object.keys(files).map((assetName) => __awaiter(this, void 0, void 0, function* () {
                const asset = yield webpack_tools_1.fileToAsset(files[assetName]);
                compilation.assets[assetName] = asset;
                webpack_tools_1.addFilesToCompilation(compilation, files[assetName]);
            }))).then(() => undefined);
        });
    }
}
exports.MixerPlugin = MixerPlugin;
