import { IPackageConfig } from '@mixer/cdk-std/dist/package';
/**
 * Prefix shown before notification lines before the JSON-encoded INotification.
 */
export declare const notificationPrefix = "miix-status:";
/**
 * Environment variable set when the plugin is run from the miix UI.
 * If set, the Notifier will emit.
 */
export declare const notificationEnv = "MIIX_UI_HOSTED";
/**
 * State of the compilation.
 */
export declare const enum CompilationState {
    Started = 0,
    Success = 1,
    Error = 2,
}
/**
 * Type of emitted notification.
 */
export declare const enum NotificationType {
    Status = 0,
    Metadata = 1,
    BundleCreated = 2,
    BundleFailed = 3,
}
/**
 * Fired when the compilation state changes.
 */
export interface IStateNotification {
    state: CompilationState;
    kind: NotificationType.Status;
}
/**
 * Fired when the project metadata changes.
 */
export interface IMetadataNotification {
    metadata: IPackageConfig;
    kind: NotificationType.Metadata;
}
/**
 * Emits that a bundle has been created. Contains the location of the
 * tarball on the filesystem.
 */
export interface IBundleCreated {
    kind: NotificationType.BundleCreated;
    location: string;
    readme: string | null;
}
/**
 * Emits that bundle creation has failed.
 */
export interface IBundleFailed {
    kind: NotificationType.BundleFailed;
    error: string;
}
export declare type Notification = IStateNotification | IMetadataNotification | IBundleCreated | IBundleFailed;
/**
 * The Notifier is loaded in development mode. It'll print a friendly blob
 * of JSON on the console that the UI will be able to read to signal
 * status of the compilation.
 *
 * Webpack compiler docs are sparse, of course, this is
 * inspired by {@link https://git.io/vxB3O}.
 */
export declare class Notifier {
    readonly isEnabled: boolean;
    constructor(isEnabled?: boolean);
    /**
     * Attaches the Notifier to compiler states.
     */
    apply(compiler: any): void;
    /**
     * Writes a notification to the console out.
     */
    printNotification(notification: Notification): void;
    /**
     * Should be called when project metadata changes, or is re-inspected.
     */
    updateMetadata(metadata: IPackageConfig): void;
    private printStatus(state);
}
/**
 * Reads the notification from the console output line. Returns null if
 * the line does not contain a notification.
 */
export declare function readNotification(line: string): Notification | null;
