"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Prefix shown before notification lines before the JSON-encoded INotification.
 */
exports.notificationPrefix = 'miix-status:';
/**
 * Environment variable set when the plugin is run from the miix UI.
 * If set, the Notifier will emit.
 */
exports.notificationEnv = 'MIIX_UI_HOSTED';
/**
 * The Notifier is loaded in development mode. It'll print a friendly blob
 * of JSON on the console that the UI will be able to read to signal
 * status of the compilation.
 *
 * Webpack compiler docs are sparse, of course, this is
 * inspired by {@link https://git.io/vxB3O}.
 */
class Notifier {
    constructor(isEnabled = !!process.env[exports.notificationEnv]) {
        this.isEnabled = isEnabled;
    }
    /**
     * Attaches the Notifier to compiler states.
     */
    apply(compiler) {
        this.printStatus(0 /* Started */);
        compiler.hooks.invalid.tap('MiixWebpackPlugin#notifier', () => {
            this.printStatus(0 /* Started */);
        });
        compiler.hooks.done.tap('MiixWebpackPlugin#notifier', (stats) => {
            if (stats.hasErrors()) {
                this.printStatus(2 /* Error */);
            }
            else {
                this.printStatus(1 /* Success */);
            }
        });
    }
    /**
     * Writes a notification to the console out.
     */
    printNotification(notification) {
        if (!this.isEnabled) {
            return;
        }
        // \n before just in case whoever came before us didn't clean up
        // after themselves, or is writing an async stream.
        process.stderr.write(`\n${exports.notificationPrefix}${JSON.stringify(notification)}\n`);
    }
    /**
     * Should be called when project metadata changes, or is re-inspected.
     */
    updateMetadata(metadata) {
        this.printNotification({ kind: 1 /* Metadata */, metadata });
    }
    printStatus(state) {
        this.printNotification({ kind: 0 /* Status */, state });
    }
}
exports.Notifier = Notifier;
/**
 * Reads the notification from the console output line. Returns null if
 * the line does not contain a notification.
 */
function readNotification(line) {
    line = line.trim();
    if (!line.startsWith(exports.notificationPrefix)) {
        return null;
    }
    return JSON.parse(line.slice(exports.notificationPrefix.length));
}
exports.readNotification = readNotification;
