"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const chai_1 = require("chai");
const path = require("path");
const rimraf = require("rimraf");
const tar = require("tar");
const webpack = require("webpack");
const fs_1 = require("fs");
const bundle_emitter_1 = require("../src/bundle-emitter");
const fs_2 = require("../src/fs");
const notifier_1 = require("../src/notifier");
function interceptStream(stream) {
    const oldWrite = stream.write;
    let data = '';
    stream.write = (d) => {
        data += d.toString();
    };
    return () => {
        stream.write = oldWrite;
        return data;
    };
}
/**
 * Runs webpack on a directory, intercepting and returning stdout/err. Yes,
 * this is a terrible function :)
 */
function packDirectory(dir) {
    return __awaiter(this, void 0, void 0, function* () {
        const restoreStdout = interceptStream(process.stdout);
        const restoreStderr = interceptStream(process.stderr);
        Object.assign(process.env, {
            [notifier_1.notificationEnv]: 1,
            [bundle_emitter_1.bundleEnv]: 1,
        });
        const context = path.join(__dirname, 'fixtures', dir);
        // tslint:disable-next-line
        const config = require(path.join(context, 'webpack.config.ts'));
        const compiler = webpack(Object.assign({}, config, { context }));
        yield fs_2.promiseCallback(callback => {
            compiler.run(callback);
        });
        const stdout = restoreStdout();
        const stderr = restoreStderr();
        const notifications = stderr
            .split('\n')
            .map(notifier_1.readNotification)
            .filter(notification => !!notification)
            .map(notification => notification);
        const tarNotification = notifications.find(n => n.kind === 2 /* BundleCreated */);
        const tarLocation = tarNotification && tarNotification.location;
        return { stdout, stderr, notifications, tarLocation };
    });
}
function listTarballFiles(filename) {
    return __awaiter(this, void 0, void 0, function* () {
        const entries = [];
        yield tar.list({
            file: filename,
            strict: true,
            onentry(entry) {
                entries.push(String(entry.path));
            },
        }, []);
        return entries.sort();
    });
}
function clearDirectory(dir) {
    rimraf.sync(path.join(__dirname, 'fixtures', dir, 'dist'));
}
describe('e2e', () => {
    describe('simple-build', () => {
        let results;
        before(() => __awaiter(this, void 0, void 0, function* () {
            results = yield packDirectory('simple-build');
        }));
        after(() => {
            clearDirectory('simple-build');
            fs_1.unlinkSync(results.tarLocation);
        });
        it('emits notifications correctly', () => __awaiter(this, void 0, void 0, function* () {
            chai_1.expect(results.notifications).to.deep.equal([
                {
                    kind: 0 /* Status */,
                    state: 0 /* Started */,
                },
                {
                    kind: 1 /* Metadata */,
                    metadata: { name: 'hello-world', version: '0.1.0', controls: {}, scenes: {} },
                },
                {
                    kind: 0 /* Status */,
                    state: 1 /* Success */,
                },
                {
                    kind: 2 /* BundleCreated */,
                    location: results.tarLocation,
                    readme: null,
                },
            ]);
        }));
        it('builds bundle outputs', () => __awaiter(this, void 0, void 0, function* () {
            chai_1.expect(yield listTarballFiles(results.tarLocation)).to.deep.equal([
                'bundle.js',
                'index.html',
                'package.json',
            ]);
        }));
    });
    describe('build-with-readme', () => {
        let results;
        before(() => __awaiter(this, void 0, void 0, function* () {
            results = yield packDirectory('build-with-readme');
        }));
        after(() => {
            clearDirectory('build-with-readme');
            fs_1.unlinkSync(results.tarLocation);
        });
        it('emits notifications correctly', () => __awaiter(this, void 0, void 0, function* () {
            chai_1.expect(results.notifications).to.deep.equal([
                {
                    kind: 0 /* Status */,
                    state: 0 /* Started */,
                },
                {
                    kind: 1 /* Metadata */,
                    metadata: { name: 'hello-world', version: '0.1.0', controls: {}, scenes: {} },
                },
                {
                    kind: 0 /* Status */,
                    state: 1 /* Success */,
                },
                {
                    kind: 2 /* BundleCreated */,
                    location: results.tarLocation,
                    readme: '<h1 id="build-with-readme">build-with-readme</h1>\n',
                },
            ]);
        }));
        it('builds bundle outputs', () => __awaiter(this, void 0, void 0, function* () {
            chai_1.expect(yield listTarballFiles(results.tarLocation)).to.deep.equal([
                'bundle.js',
                'index.html',
                'package.json',
                'readme.md',
            ]);
        }));
    });
    describe('invalid build', () => {
        let results;
        before(() => __awaiter(this, void 0, void 0, function* () {
            results = yield packDirectory('invalid-build');
        }));
        it('emits notifications correctly', () => __awaiter(this, void 0, void 0, function* () {
            chai_1.expect(results.notifications).to.deep.equal([
                {
                    kind: 0 /* Status */,
                    state: 0 /* Started */,
                },
                {
                    kind: 0 /* Status */,
                    state: 2 /* Error */,
                },
            ]);
        }));
    });
});
